/* Funkbots 1.15: Copyright 1998  Jonathan Sharkey. Please read reame.txt
for important info */

#include "fb.h"

struct var_struct var;
struct robot_struct robot[256];
struct item_struct item[256];
struct shot_struct shot[1000];
struct expl_struct expl[1000];
struct cover_it_struct cover_it[1000];

DATAFILE *gamedata;
BITMAP *bat_scr;

int main(int argc, char *argv[])
{
  char battle_file[160] = "default.btl", errmsg[210];
  time_t seed;

  /* if there, set argument as battle file, other wise keep as default.btl */
  if (argc > 1)
    strcpy(battle_file, argv[1]);

  /* set up random number generator */
  seed = time(NULL);
  srandom(seed);

  /* try to load file. returns string containing "" is successfull, an error
  message other wise */
  errmsg[0] = 0;
  load_battle_file(battle_file, errmsg);
  if (errmsg[0]) {
    puts(errmsg);
    return(255);
  }

  init();     /* set up allegro */
  do_titlescreen();
  do_battle();

  /* free allocated memory */
  quit();

  return(0);
}

void init(void)
{
  ISHORT i;

  allegro_init();
  install_timer();
  install_mouse();
  install_keyboard();
  set_gfx_mode(var.video, var.res_x, var.res_y, var.res_x, var.res_y);
  reserve_voices(32, 0);
  install_sound(var.sound, MIDI_NONE, "");
  set_volume(255, 0);

  /* Load graphics and sound */
  gamedata = load_datafile("FB.DAT");
  bat_scr = create_bitmap(var.res_x, var.res_y);

  /* If robots don't use custom bitmap assign default */
  for (i = 0; i < var.total_robots; i++)
    if (!robot[i].image)
      robot[i].image = gamedata[ROBOT].dat;
}

void do_titlescreen(void)
{
  char string[20];
  ISHORT pos = 0, i;

  set_palette(black_palette);

  stretch_blit(gamedata[START_UP].dat, screen, 0, 0, 281, 200,
    0, 0, var.res_x, var.res_y);
  fade_in(gamedata[PAL_START].dat, 8);
  rest(2000);
  fade_out(8);

  rectfill(screen, 0, 0, var.res_x - 1, var.res_y - 1, 87);
  blit(gamedata[TITLE].dat, screen, 0, 0, var.res_x / 2 - 293, 10, 586, 70);

  text_mode(87);
  sprintf(string, "Total Robots: %i", var.total_robots);
  textout(screen, font, string, 20, 100, 0);
  sprintf(string, "Robot Energy: %i", var.robot_energy);
  textout(screen, font, string, 20, 110, 0);
  sprintf(string, "Item Delay:   %i", var.item_delay);
  textout(screen, font, string, 20, 125, 0);
  sprintf(string, "Normal Shot:  %i", var.item_chance[NORMAL]);
  textout(screen, font, string, 20, 135, 0);
  sprintf(string, "Power Shot:   %i", var.item_chance[POWER]);
  textout(screen, font, string, 20, 145, 0);
  sprintf(string, "Seek Shot:    %i", var.item_chance[SEEKS]);
  textout(screen, font, string, 20, 155, 0);
  sprintf(string, "Power Seek:   %i", var.item_chance[PSEEK]);
  textout(screen, font, string, 20, 165, 0);
  sprintf(string, "Mine Shot:    %i", var.item_chance[MINES]);
  textout(screen, font, string, 20, 175, 0);
  sprintf(string, "Power Mine:   %i", var.item_chance[PMINE]);
  textout(screen, font, string, 20, 185, 0);
  sprintf(string, "Energy 1:     %i", var.item_chance[ENERGY1]);
  textout(screen, font, string, 20, 195, 0);
  sprintf(string, "Energy 2:     %i", var.item_chance[ENERGY2]);
  textout(screen, font, string, 20, 205, 0);
  sprintf(string, "Energy 5:     %i", var.item_chance[ENERGY5]);
  textout(screen, font, string, 20, 215, 0);
  sprintf(string, "Game Speed:   %i", var.game_speed);
  textout(screen, font, string, 20, 230, 0);

  for (i = 0; i < var.total_robots && i < (var.res_y - 100) / 42; i++) {
    sprintf(string, "Robot %i: %s", i, robot[i].name);
    textout(screen, font, string, 300, 100 + i * 40, 0);
    sprintf(string, "X: %i", robot[i].x);
    textout(screen, font, string, 300, 110 + i * 40, 0);
    sprintf(string, "Y: %i", robot[i].y);
    textout(screen, font, string, 300, 120 + i * 40, 0);
  }

  fade_in(gamedata[PAL].dat, 8);

  while(!key[KEY_ENTER]) {
    if (key[KEY_UP] && pos > 0) {
      pos--;
      blit(screen, screen, 300, 100, 300, 140, 200, var.res_y - 200);
      sprintf(string, "Robot %i: %s", pos, robot[pos].name);
      textout(screen, font, string, 300, 100, 0);
      sprintf(string, "X: %i", robot[pos].x);
      textout(screen, font, string, 300, 110, 0);
      sprintf(string, "Y: %i", robot[pos].y);
      textout(screen, font, string, 300, 120, 0);
      while (key[KEY_UP]);
    }
    if (key[KEY_DOWN] && pos < var.total_robots - (var.res_y - 100) / 42) {
      pos++;
      blit(screen, screen, 300, 140, 300, 100, 200, var.res_y - 140);
      sprintf(string, "Robot %i: %s", pos + (var.res_y - 100) / 42 - 1,
        robot[pos + (var.res_y - 100) / 42 - 1].name);
      textout(screen, font, string, 300,
        100 + ((var.res_y - 100) / 42 - 1) * 40, 0);
      sprintf(string, "X: %i", robot[pos + (var.res_y - 100) / 42 - 1].x);
      textout(screen, font, string, 300,
        110 + ((var.res_y - 100) / 42 - 1) * 40, 0);
      sprintf(string, "Y: %i", robot[pos + (var.res_y - 100) / 42 - 1].y);
      textout(screen, font, string, 300,
        120 + ((var.res_y - 100) / 42 - 1) * 40, 0);
      while (key[KEY_DOWN]);
    }
  }

  fade_out(8);
}

void quit(void)
{
  ISHORT i;

  for (i = 0; i < var.total_robots; i++) {
    free(robot[i].code);
    if (!robot[i].image || robot[i].image != gamedata[ROBOT].dat)
      destroy_bitmap(robot[i].image);
  }

  unload_datafile(gamedata);
  destroy_bitmap(bat_scr);
  remove_int(counter);
  if (var.movie_s)
    pack_fclose(var.movie_f);
}
