/* Funkbots 1.15: Copyright 1998  Jonathan Sharkey. Please read reame.txt
for important info */

#include "fb.h"

void load_battle_file(char *battle_file, char *errmsg)
{
  FILE *file;
  char fileline[81];
  ISHORT i, j;
  PALETTE pal;

  /* return error if file not found */
  if (!__file_exists(battle_file)) {
    sprintf(errmsg, "\nError: %s doesn't exist", battle_file);
    return;
  }

  /* set all stats to 0 */
  for (i = 0; i < 256; i++) {
    memset(robot[i].data, 0, sizeof(float) * 1000);
    memset(robot[i].ammo, 0, sizeof(ISHORT) * 6);
    robot[i].shot_count = 0;
    robot[i].total_hits = 0;
    memset(robot[i].hits, 0, sizeof(ISHORT) * 256);
    robot[i].kills = 0;
    robot[i].debug = 0;
    robot[i].code_size = 0;
    robot[i].user_ctrl = 0;
  }

  /* set up default options */
  var.video = GFX_AUTODETECT;
  var.sound = DIGI_AUTODETECT;
  var.game_speed = 60;
  var.total_robots = 0;
  var.robot_energy = 20;
  var.total_items = 0;
  var.item_delay = 60;
  var.res_x = 640;
  var.res_y = 480;
  var.area_x = 538;
  var.area_y = 480;

  for (i = 0; i < 9; i++) {
    var.num_item[i] = 0;
    var.item_chance[i] = 10;
  }
  var.item_chance[ENERGY5] = 3;

  /* read battle file and set up game */
  file = fopen(battle_file, "rt");
  while(1) {
    if (feof(file))
      break;

    fgets(fileline, 80, file);   /* read in line from file */

    /* only use what's up to the first space, or end of line */
    for (i = 0; i < strlen(fileline); i++)
      if (fileline[i] == ' ' || fileline[i] == '\n')
        fileline[i] = 0;
    /* read up to first '=', then use it as needed */
    for (i = 0; i < strlen(fileline); i++) {
      if (fileline[i] == '=')  {
        fileline[i] = 0;
        if (!stricmp(fileline, "ROBOT")) {
          load_robot(&fileline[i + 1], errmsg);
          if (errmsg[0])  /* quit if load_robot returns an error */
            return;
          break;
        }
        if (!stricmp(fileline, "VIDEO"))
          var.video = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "SOUND"))
          var.sound = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "GAME.SPEED"))
          var.game_speed = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "ROBOT.ENERGY"))
          var.robot_energy = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "ITEM.DELAY"))
          var.item_delay = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "NORMAL"))
          var.item_chance[NORMAL] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "POWER"))
          var.item_chance[POWER] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "SEEKING"))
          var.item_chance[SEEKS] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "POWER.SEEK"))
          var.item_chance[PSEEK] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "MINE"))
          var.item_chance[MINES] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "POWER.MINE"))
          var.item_chance[PMINE] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "ENERGY.1"))
          var.item_chance[ENERGY1] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "ENERGY.2"))
          var.item_chance[ENERGY2] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "ENERGY.5"))
          var.item_chance[ENERGY5] = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "REC.MOVIE")) {
          var.movie_s = GREC;
          strcpy(var.movie_name, &fileline[i + 1]);
        }
        if (!stricmp(fileline, "PLAY.MOVIE")) {
          var.movie_s = GPLAY;
          strcpy(var.movie_name, &fileline[i + 1]);
        }
        if (!stricmp(fileline, "IMAGE"))
          robot[var.total_robots - 1].image =
          load_bitmap(&fileline[i + 1], pal);
        if (!stricmp(fileline, "AREA.X"))
          var.area_x = atoi(&fileline[i + 1]);
        if (!stricmp(fileline, "AREA.Y"))
          var.area_y = atoi(&fileline[i + 1]);

        if (!stricmp(fileline, "RES")) {
          j = atoi(&fileline[i + 1]);
          if (j == 0) {
            var.res_x = 640;
            var.res_y = 480;
          }
          if (j == 1) {
            var.res_x = 800;
            var.res_y = 600;
          }
          if (j == 2) {
            var.res_x = 1024;
            var.res_y = 768;
          }
          if (j == 3) {
            var.res_x = 1280;
            var.res_y = 1024;
          }
          if (j == 4) {
            var.res_x = 1600;
            var.res_y = 1200;
          }
        }
      }
    }
  }
  for (i = 0; i < var.total_robots; i++) {
    robot[i].x = randnum(var.area_x);
    robot[i].y = randnum(var.area_y);
    robot[i].energy = var.robot_energy;
  }

  if (var.movie_s != NONE)
    setup_movie(errmsg);
  for (i = 0; i < 9; i++) {
    for (j = var.num_chance; j < var.num_chance + var.item_chance[i]; j++)
      var.chance[j] = i;
      var.num_chance += var.item_chance[i];
  }

  fclose(file);
}

void load_robot(char *robot_file, char *errmsg)
{
  ISHORT i;

  if (stricmp(robot_file, "USERCTRL")) {
    if (!__file_exists(robot_file)) {
      sprintf(errmsg, "\nError: Robot file not found, %s", robot_file);
      return;
    }

    robot[var.total_robots].code_size = code_size(robot_file) + 1;
    robot[var.total_robots].code = (ISHORT*)
      malloc((ISHORT)robot[var.total_robots].code_size * sizeof(ISHORT));
    if (robot[var.total_robots].code == NULL) {
      strcpy(errmsg, "Error allocating memory!");
      return;
    }
    load_code(robot_file);
    robot[var.total_robots].code[robot[var.total_robots].code_size - 1] = 0;
  }
  else
    robot[i].user_ctrl = TRUE;

  strcpy(robot[var.total_robots].name, robot_file);
  robot[var.total_robots].alive = TRUE;
  for (i = 0; i < 9; i++)
    if (robot[var.total_robots].name[i] == '.')
      robot[var.total_robots].name[i] = 0;
  robot[var.total_robots].name[9] = 0;

  var.total_robots++;
  var.robots_alive++;
}

/* caluculate number of bytes needed for compiled robot code */
int code_size(char *robot_file)
{
  FILE *file;
  char fileline[200], command[200], definew[200][200], definer[200][200];
  char *pos, temp[400];
  ISHORT num_d = 0;

  char *vars1[66] =
  {"ROBOT.X", "ROBOT.Y", "DATA", "ROBOT.ENERGY", "ROBOT.AMMO",
  "ROBOTS.ALIVE", "TOTAL.ROBOTS", "TOTAL.SHOTS",
  "ROBOT.ALIVE", "ITEM.DELAY", "NUM.ITEMS",
  "ITEM.CHANCE", "ITEM.X", "ITEM.Y", "ITEM.TYPE", "NUM.SHOTS",
  "SHOT.X", "SHOT.Y", "SHOT.SPEED.X", "SHOT.SPEED.Y", "SHOT.TYPE", "NORMAL",
  "POWER", "SEEKING", "POWER.SEEK", "MINE", "POWER.MINE", "ENERGY.1",
  "ENERGY.2", "ENERGY.5", "+", "-", "/", "*", "%", "=", ">", ">=", "<=",
  "<>", "IF", "ENDIF", "DO", "BREAK", "LOOP", "SHOOT", "ME", "]", "(", ")",
  "RAND", "&&", "||", "<", "ROBOT.TOTAL.HITS", "ROBOT.HITS", "ROBOT.KILLS",
  "SHOT.ROBOT", "AREA.X", "AREA.Y", "DEBUG", "AREA.X", "AREA.Y", "SQRT",
  "^", "ELSE"};

  int size = 0;
  ISHORT i, k;

  file = fopen(robot_file, "rt");

  while(1) {
    skipper:

    if (feof(file))
      break;

    fgets(fileline, 80, file);

    strupr(fileline);
    format_string(fileline);
    sprintf(temp, " %s ", fileline);
    strcpy(fileline, temp);

    do_again2:
    for (i = 0; i < num_d; i++) {
      pos = (char *) strstr(fileline, definew[i]);
      if (pos && pos[-1] == ' ' && pos[strlen(definew[i])] == ' ') {
        *pos = 0;
        strcpy(temp, fileline);
        strcat(temp, definer[i]);
        strcat(temp, pos + strlen(definew[i]));
        sprintf(fileline, " %s", temp);
        goto do_again2;
      }
    }

    if (!format_string(fileline))
      goto skipper;

    while (get_command(command, fileline)[0]) {
      k = 0;

      if (!strcmp(command, ";"))
        break;

      if (!strcmp(command, "DEFINE") && num_d < 200) {
        strcpy(definew[num_d], get_command(command, fileline));
        strcpy(definer[num_d], fileline);
        /* get rid of comment after define */
        for (i = 0; i < strlen(definer[num_d]); i++)
          if (definer[num_d][i] == ';')
            definer[num_d][i] = 0;

        strcpy(fileline, " ");
        num_d++;
        goto skipper;
      }

      for (i = 0; i < 66; i++)
        if (!strcmp(command, vars1[i])) {
          size++;
          k = 1;
        }
      if (!k && command[0] != 0) {
        /* make sure it's a number and not a misspelled word */
        for (k = 0; k < strlen(command); k++)
          if ((command[k] < '0' || command[k] > '9') &&
          command[k] != '.' && command[k] != '-') {
            printf("\nError in %s: %s unknown command.\n", robot_file, command);
            quit();
            exit(1);
          }
        size += 3;
      }
    }
  }
  fclose(file);

  return(size);
}

ISHORT format_string(char *fileline)
{
  ISHORT k = 0, i, j, length;

  for (i = 0; i < strlen(fileline); i++) {
    if (fileline[i] == '[' || fileline[i] == '\n')
      fileline[i] = ' ';
    if (fileline[i] != ' ')
      k++;
  }

  if (!k)
    return(0);

  /* put spaces around ;, [, (, and ) */

  for (i = 0; i < strlen(fileline); i++) {
    length = strlen(fileline);
    if (fileline[i] == ']' || fileline[i] == '(' || fileline[i] == ')' ||
    fileline[i] == ';') {
      for (j = length; j >= i; j--)
        fileline[j + 1] = fileline[j];
      for (j = length + 1; j >= i + 2; j--)
        fileline[j + 1] = fileline[j];
      fileline[i] = ' ';
      fileline[i + 2] = ' ';
      i += 2;
    }
  }

  if (fileline[strlen(fileline) - 1] != ' ') {
    fileline[strlen(fileline) + 1] = 0;
    fileline[strlen(fileline)] = ' ';
  }

  for (i = strlen(fileline) - 1; i > 0; i--)
    if (fileline[i] != ' ' && fileline[i] != 9) {
      fileline[i + 2] = 0;
      goto breaker;
    }
  breaker:

  return(1);
}

char *get_command(char *command, char *fileline)
{
  ISHORT i, j, k;

  for (i = 0; i < strlen(fileline) &&
  (fileline[i] == ' ' || fileline[i] == 9); i++);

  if (strlen(fileline) == 0) {
    command[0] = 0;
    return(command);
  }

  for (j = i; j < strlen(fileline); j++) {
    if (fileline[j] == ' ' || fileline[j] == 9 || fileline[j] == 0) {
      for (k = i; k < j; k++)
        command[k - i] = fileline[k];
      command[j - i] = 0;
      strcpy(fileline, &fileline[j + 1]);
      return(command);
    }
  }
}

void load_code(char *robot_file)
{
  char *vars1[66] =
  {"ROBOT.X", "ROBOT.Y", "DATA", "ROBOT.ENERGY", "ROBOT.AMMO",
  "ROBOTS.ALIVE", "TOTAL.ROBOTS", "TOTAL.SHOTS",
  "ROBOT.ALIVE", "ITEM.DELAY", "NUM.ITEMS",
  "ITEM.CHANCE", "ITEM.X", "ITEM.Y", "ITEM.TYPE", "NUM.SHOTS",
  "SHOT.X", "SHOT.Y", "SHOT.SPEED.X", "SHOT.SPEED.Y", "SHOT.TYPE", "NORMAL",
  "POWER", "SEEKING", "POWER.SEEK", "MINE", "POWER.MINE", "ENERGY.1",
  "ENERGY.2", "ENERGY.5", "+", "-", "/", "*", "%", "=", ">", ">=", "<=",
  "<>", "IF", "ENDIF", "DO", "BREAK", "LOOP", "SHOOT", "ME", "]", "(", ")",
  "RAND", "&&", "||", "<", "ROBOT.TOTAL.HITS", "ROBOT.HITS", "ROBOT.KILLS",
  "SHOT.ROBOT", "AREA.X", "AREA.Y", "DEBUG", "AREA.X", "AREA.Y", "SQRT",
  "^", "ELSE"};
  char fileline[400], command[400], definew[200][200], definer[200][200];
  char *pos, *pos2, temp[400];
  FILE *file;
  ISHORT i, k, num_d = 0;
  int code_pos = 0;
  float *val;

  file = fopen(robot_file, "rt");

  while(1) {
    skipper2:
    if (feof(file))
      break;

    fgets(fileline, 80, file);

    strupr(fileline);
    format_string(fileline);
    sprintf(temp, " %s ", fileline);
    strcpy(fileline, temp);

    do_again2:
    for (i = 0; i < num_d; i++) {
      pos = (char *) strstr(fileline, definew[i]);
      if (pos && pos[-1] == ' ' && pos[strlen(definew[i])] == ' ') {
        *pos = 0;
        strcpy(temp, fileline);
        strcat(temp, definer[i]);
        strcat(temp, pos + strlen(definew[i]));
        sprintf(fileline, " %s", temp);
        goto do_again2;
      }
    }

    if (!format_string(fileline))
      goto skipper2;


    marker:
    while (get_command(command, fileline)[0]) {
      k = 0;

      if (!strcmp(command, ";"))
        break;

      if (!strcmp(command, "DEFINE") && num_d < 200) {
        strcpy(definew[num_d], get_command(command, fileline));
        strcpy(definer[num_d], fileline);
        /* get rid of comment after define */
        for (i = 0; i < strlen(definer[num_d]); i++)
          if (definer[num_d][i] == ';')
            definer[num_d][i] = 0;

        strcpy(fileline, " ");
        num_d++;
        goto skipper2;
      }

      for (i = 0; i < 66; i++)
        if (!strcmp(command, vars1[i])) {
          robot[var.total_robots].code[code_pos] = i + 1;
          code_pos++;
          k = 1;
        }
      if (!k && command[0] != 0) {
        robot[var.total_robots].code[code_pos] = 0;
        val = (float *)&robot[var.total_robots].code[code_pos + 1];
        *val = (float)atof(command);
        code_pos += 3;
      }
    }
  }
  fclose(file);
}

int randnum(int max)
{
  ISHORT number;

  number = random() % max;
  if (var.movie_s == REC)
    pack_fwrite(&number, sizeof(ISHORT), var.movie_f);
  if (var.movie_s == PLAY && !pack_feof(var.movie_f))
    pack_fread(&number, sizeof(ISHORT), var.movie_f);
  return((int)number);
}

void setup_movie(char *errmsg)
{
  ISHORT i, t = 1, f = 0;

  if (var.movie_s == GREC) {
    var.movie_f = pack_fopen(var.movie_name, F_WRITE_PACKED);
    if (!var.movie_f) {
      sprintf(errmsg, "Error opening %s for writing.", var.movie_name);
      return;
    }
    pack_fwrite(&var.area_x, sizeof(ISHORT), var.movie_f);
    pack_fwrite(&var.area_y, sizeof(ISHORT), var.movie_f);
    pack_fwrite(&var.total_robots, sizeof(ISHORT), var.movie_f);
    pack_fwrite(&var.robot_energy, sizeof(ISHORT), var.movie_f);
    pack_fwrite(&var.item_delay, sizeof(ISHORT), var.movie_f);
    pack_fwrite(&var.item_chance, sizeof(ISHORT) * 9, var.movie_f);
    for (i = 0; i < var.total_robots; i++) {
      if (robot[i].image == NULL)
        pack_fwrite(&t, sizeof(ISHORT), var.movie_f);
      else {
        pack_fwrite(&f, sizeof(ISHORT), var.movie_f);
        pack_fwrite(robot[i].image->dat, sizeof(char) * 30 * 42, var.movie_f);
      }
      pack_fwrite(&robot[i].name, sizeof(char) * 9, var.movie_f);
      pack_fwrite(&robot[i].x, sizeof(ISHORT), var.movie_f);
      pack_fwrite(&robot[i].y, sizeof(ISHORT), var.movie_f);
      pack_fwrite(&robot[i].code_size, sizeof(int), var.movie_f);
      pack_fwrite(robot[i].code, sizeof(ISHORT) * robot[i].code_size,
      var.movie_f);
    }
    var.movie_s = REC;
  }

  if (var.movie_s == GPLAY) {
    for (i = 0; i < var.total_robots; i++) {
      free(robot[i].code);
      if (robot[i].image != gamedata[ROBOT].dat)
        destroy_bitmap(robot[i].image);
    }
    for (i = 0; i < 256; i++) {
      robot[i].x = 0;
      robot[i].y = 0;
      robot[i].energy = 0;
    }
    var.movie_f = pack_fopen(var.movie_name, F_READ_PACKED);
    if (!var.movie_f) {
      sprintf(errmsg, "Error opening %s for reading.", var.movie_name);
      return;
    }
    pack_fread(&var.area_x, sizeof(ISHORT), var.movie_f);
    pack_fread(&var.area_y, sizeof(ISHORT), var.movie_f);
    pack_fread(&var.total_robots, sizeof(ISHORT), var.movie_f);
    pack_fread(&var.robot_energy, sizeof(ISHORT), var.movie_f);
    pack_fread(&var.item_delay, sizeof(ISHORT), var.movie_f);
    pack_fread(&var.item_chance, sizeof(ISHORT) * 9, var.movie_f);
    for (i = 0; i < var.total_robots; i++) {
      pack_fread(&t, sizeof(ISHORT), var.movie_f);
      if (t == 1)
        robot[i].image = NULL;
      else {
        robot[i].image = create_bitmap(30, 42);
        if (!robot[i].image) {
          quit();
          exit(1);
        }
        pack_fread(robot[i].image->dat, sizeof(char) * 30 * 42, var.movie_f);
      }

      pack_fread(&robot[i].name, sizeof(char) * 9, var.movie_f);
      pack_fread(&robot[i].x, sizeof(ISHORT), var.movie_f);
      pack_fread(&robot[i].y, sizeof(ISHORT), var.movie_f);
      pack_fread(&robot[i].code_size, sizeof(int), var.movie_f);
      robot[i].code = (ISHORT *) malloc(robot[i].code_size * sizeof(ISHORT));
      if (!robot[i].code) {
        strcpy(errmsg, "Memory allocation error.");
        return;
      }
      pack_fread(robot[i].code, sizeof(ISHORT) * robot[i].code_size,
      var.movie_f);
      robot[i].alive = TRUE;
    }
    var.robots_alive = var.total_robots;
    var.movie_s = PLAY;
  }
}
